type int = Int.t

functor Document (S: DOCUMENT_STRUCTS): DOCUMENT =
struct

open S

datatype z = datatype Color.t
   
datatype tag = B | EM | I | PL | S | TT | U
             | Size of int | Color of Color.t

val tagToStartEndStrings =
   fn B => ("<B>","</B>")
    | EM => ("<EM>","</EM>")
    | I => ("<I>","</I>")
    | PL => ("<PL>","</PL>")
    | S => ("<S>","</S>")
    | TT => ("<TT>","</TT>")
    | U => ("<U>","</U>")
    | Size n =>
	 (case n of
	     0 => ("<0>","</0>")
	   | 1 => ("<1>","</1>")
	   | 2 => ("<2>","</2>")
	   | 3 => ("<3>","</3>")
	   | 4 => ("<4>","</4>")
	   | 5 => ("<5>","</5>")
	   | 6 => ("<6>","</6>")
	   | 7 => ("<7>","</7>")
	   | 8 => ("<8>","</8>")
	   | 9 => ("<9>","</9>")
	   | _ => Error.bug "strange size tag")
    | Color c =>
	 let
	    datatype z = datatype Color.t
	 in
	    case c of
	       Black => ("<k>","</k>")
	     | Blue => ("<b>","</b>")
	     | Cyan => ("<c>","</c>")
	     | Green => ("<g>","</g>")
	     | Magenta => ("<m>","</m>")
	     | Red => ("<r>","</r>")
	     | Unknown => Error.bug "tried to print tag with Unknown color"
	     | White => ("<w>","</w>")
	     | Yellow => ("<y>","</y>")
	 end

val startTagToString = #1 o tagToStartEndStrings
val endTagToString = #2 o tagToStartEndStrings

val tagToStartSize : tag -> int =
   fn B => 3
    | EM => 4
    | I => 3
    | PL => 4
    | S => 3
    | TT => 4
    | U => 3
    | Size n => 3
    | Color c => 3

datatype t = Doc of item list
     and item = Text of string
              | Tagged of tag * t

fun output (doc, os)
  = let
      val rec output'
	= fn [] => ()
	   | i::is => (outputItem' i ; output' is)
      and outputItem'
	= fn Text s => Out.output (os, s)
	   | Tagged (t, Doc is) 
	   => let
		val (st,et) = tagToStartEndStrings t
	      in
		Out.output (os, st) ;
		output' is ;
		Out.output (os, et)
	      end
      val Doc is = doc
    in
      output' is
    end

fun toString doc
  = let
      val rec toString'
	= fn is => concat (List.map(is, toStringItem'))
      and toStringItem'
	= fn Text s => s
	   | Tagged (t, Doc is) 
	   => let
		val (st,et) = tagToStartEndStrings t
	      in
		concat [st, toString' is, et]
	      end
      val Doc is = doc
    in
      toString' is
    end

fun size doc
  = let
      val rec size'
	= fn (is, size) => List.fold(is, size, fn (i,size) => sizeItem' (i, size))
      and sizeItem'
	= fn (Text s, size) => size + (String.size s)
	   | (Tagged (t, Doc is), size)
	   => let
		val s = tagToStartSize t
	      in
		size' (is, size + 2 * s + 1)
	      end
      val Doc is = doc
    in
      size' (is, 0)
    end

end
