type int = Int.t

functor GenRandom (S: GENRANDOM_STRUCTS): GENRANDOM =
struct

open S

structure Color = Document.Color

datatype dowhat1 = Dw1Char | Dw1Start | Dw1End

val whitespacep = 0.01

fun randchar () =
    let val r = Random.real ()
    in  if r<whitespacep then
	    case Random.intRange(9,11)() of
		11 => Char.fromInt 13
	      | n => Char.fromInt n
	else
	    let val c = Random.intRange(32,124)()
	    in if c<60 then Char.fromInt c
	       else if c=60 then Char.fromInt 61
	       else Char.fromInt (c+2)
	    end
    end

fun randcolour () =
    case Random.intRange(0,7)() of
	0 => Color.Black
      | 1 => Color.Blue
      | 2 => Color.Cyan
      | 3 => Color.Green
      | 4 => Color.Magenta
      | 5 => Color.Red
      | 6 => Color.White
      | 7 => Color.Yellow
      | _ => Error.bug "bad random"

fun generate1 (size : int) =
    let val charp = 0.8
	val startp = 0.9
	val bp = 0.05
	val emp = 0.1
	val ip = 0.15
	val plp = 0.20
	val sp = 0.25
	val ttp = 0.3
	val up = 0.35
	val sizep = 0.7
	type stk = (Document.tag option * Document.item list) list
	fun pushChar (stk:stk,c) :stk =
	    case stk of 
		(tag,[])::stk => (tag,[Document.Text (String.fromChar c)])::stk
	      | (tag,(Document.Text t)::is)::stk =>
		(tag,(Document.Text (t^(String.fromChar c)))::is)::stk
	      | (tag,(is as (Document.Tagged (_,_))::_))::stk =>
		(tag,(Document.Text (String.fromChar c))::is)::stk
	      | [] => Error.bug "bad stack"
	fun pushTag (stk:stk,t) :stk = (SOME t,[])::stk
	fun popStk (stk:stk) :stk =
	    case stk of
		(SOME t,is)::(t',is')::stk' =>
		(t',(Document.Tagged (t,Document.Doc (List.rev is)))::is')::
		stk'
	      (* This case just to make the random stuff easier *)
	      | [(NONE,_)] => stk
	      | _ => Error.bug "bad stack"
	fun unwind stk =
	    case stk of
	        [(NONE,is)] => Document.Doc (List.rev is)
	      | _::_ => unwind (popStk stk)
	      | _ => Error.bug "bad stack"
	fun randdowhat () =
	    let val r = Random.real ()
	    in  if r<charp then Dw1Char
		else if r<startp then Dw1Start
		else Dw1End
	    end
	fun randstart () =
	    let val r = Random.real ()
	    in if r<bp then Document.B
	       else if r<emp then Document.EM
	       else if r<ip then Document.I
	       else if r<plp then Document.PL
	       else if r<sp then Document.S
	       else if r<ttp then Document.TT
	       else if r<up then Document.U
	       else if r<sizep then Document.Size (Random.intRange(0,9)())
	       else Document.Color (randcolour ())
	    end
	fun loop (stk, count) =
	    if count < size then
		case randdowhat () of
		    Dw1Char => loop (pushChar (stk,randchar ()),count+1)
		  | Dw1Start => loop (pushTag (stk,randstart ()),count)
		  | Dw1End => loop (popStk stk, count)
	    else
		unwind stk
    in  loop ([(NONE,[])],0)
    end

fun generate2 (size : int) =
    let val unknownp = 0.01
	val d = { c = #"a",
		  p = { b=false, color=Color.Black, i=false, em=false,
			tt=false, s=false, size=(SOME 0), u=0 } }
	val i = Meaning.toInternal d
	val is = Array.array (size,i)
	fun loop (count,prev) =
	    if count<size then
		let val c = randchar ()
		    val b = Random.bool ()
		    val r = Random.real ()
		    val color =
			if r<unknownp then Color.Unknown else randcolour ()
		    val i = Random.bool ()
		    val s = Random.bool ()
		    val em = if s then false else Random.bool ()
		    val tt = Random.bool ()
		    val r = Random.real ()
		    val size = 
			if r<unknownp then NONE
			else SOME (Random.intRange(0,9)())
		    val u = Random.intRange (0,3)()
		    val color' = if u=0 then Color.White else color
		    val d =
			if Char.isSpace c then
			    { c=if tt then c else #" ",
			      p={ b=false, color=color', i=false, em=false,
				  tt=tt, s=false, size=size, u=u } }
			else
			    { c=c,
			      p={ b=b, color=color, i=i, em=em, tt=tt, s=s,
				  size=size, u=u } }
		    val i = Meaning.toInternal d
		in  if Char.isSpace c andalso Char.isSpace (#c prev) andalso
		       (#p d) = (#p prev) andalso (not tt) then 
			loop (count,prev)
		    else
			(Array.update(is,count,i); loop (count+1, d))
		end
	    else
		()
    in  (loop (0, d); is)
    end
	
end
