type int = Int.t

functor Peephole(S: PEEPHOLE_STRUCTS): PEEPHOLE =
struct

open S
structure Color = Document.Color
structure Proprec = Meaning.Proprec
val update = Proprec.update

datatype z = datatype Color.t
datatype z = datatype Document.tag

local
  val cast_color : Document.Color.t -> Int.t
    = fn Black => 0
       | Blue => 1
       | Cyan => 2
       | Green => 3
       | Magenta => 4
       | Red => 5
       | White => 6
       | Yellow => 7
       | Unknown => Error.bug "Peephole:: cast_color"

  val cast_tag : Document.tag -> Int.t * Int.t
    = fn B => (0,0)
       | EM => (0,1)
       | I => (0,2)
       | PL => (0,3)
       | S => (0,4)
       | TT => (0,5)
       | U => (0,6)
       | Size i => (1,i)
       | Color c => (2,cast_color c)
in
  fun compare_tag (t1, t2)
    = let
	val (a1,b1) = cast_tag t1
	val (a2,b2) = cast_tag t2
      in
	if a1 = a2
	  then Int.compare (b1,b2)
	  else Int.compare (a1,a2)
      end
end

datatype z = datatype Document.t
datatype z = datatype Document.item

fun flatten(is,ctxt)
  = let
      val rec flatten'
	= fn ([],ac) => ac
	   | ((Text "")::is,ac) 
	   => flatten' ((List.rev ac) @ is,[])
	   | ((Text s1)::(Text s2)::is,ac)
	   => flatten' ((Text (s2^s1))::is,ac)
	   | ((i1 as Text s1)::is,ac)
	   => flatten' (is, i1::ac)
	   | ((Tagged (t1, Doc []))::is,ac)
	   => flatten' (is,ac)
	   | ((i1 as Tagged (t1, Doc is1))::
	      (i2 as Tagged (t2, Doc is2))::is,ac)
	   => if t1 = t2
		then let
		       val ctxt = Proprec.update(ctxt, t1)
		       val is' = List.rev(is2 @ is1)
		       val is' = flatten(is',ctxt)
		     in
		       flatten'((Tagged(t1, Doc is'))::is,ac)
		     end
		else if Proprec.equals(ctxt, update(ctxt, t1))
		       then flatten'((List.rev is1)@(i2::is),ac)
		       else flatten'(i2::is,i1::ac)
	   | ((i1 as Tagged (t1, Doc is1))::is,ac)
	   => let
		fun default ()
		  = if Proprec.equals(ctxt, update(ctxt, t1))
		      then flatten'((List.rev is1)@is,ac)
		      else flatten'(is,i1::ac)
		fun default' ()
		  = flatten'((List.rev is1)@is,ac)
		fun defaultSize'' ()
		  = case is1
		      of (i' as (Tagged (Size _, Doc is1')))::is1
		       => flatten'((Tagged (t1, Doc is1))::i'::is,ac)
		       | _ => default ()
		fun defaultColor'' ()
		  = case is1
		      of (i' as (Tagged (Color _, Doc is1')))::is1
		       => flatten'((Tagged (t1, Doc is1))::i'::is,ac)
		       | _ => default ()
	      in
		case t1
		  of Size _
		   => if List.forall(is1,
				     fn Tagged (Size _, _) => true
				      | _ => false)
			then default' ()
			else defaultSize'' ()
		   | Color _
		   => if List.forall(is1,
				     fn Tagged (Color _, _) => true
				      | _ => false)
			then default' ()
			else defaultColor'' ()
		   | _ => default ()
	      end
    in
      flatten'(is,[])
    end

fun peephole d 
  = let
      type foldState = {lastCtxt: Proprec.t option}
      type treeState = {ctxt: Proprec.t}
      fun setLastCtxt({lastCtxt}, lastCtxt') = {lastCtxt = lastCtxt'}
      fun setCtxt({ctxt}, ctxt') = {ctxt = ctxt'}

      val rec walk : Document.t * treeState * foldState -> Document.t * foldState
	= fn (Doc is, treeState, foldState)
	   => let
		val (is, foldState)
		  = List.fold
		    (is,
		     ([],foldState),
		     fn (i,(is,foldState)) 
		      => let
			   val (i,foldState) = walkItem (i, treeState, foldState)
			 in
			   (i::is,foldState)
			 end)
		val ctxt = #ctxt treeState
		val is = flatten(is, ctxt)
	      in
		(Doc is, foldState)
	      end
      and walkItem : Document.item * treeState * foldState -> Document.item * foldState
	= fn (Text s, treeState, foldState)
	   => if #tt (#ctxt treeState)
		then let
		       val ctxt = #ctxt treeState
		       val lastCtxt = (if Char.isSpace (String.last s)
					 then SOME ctxt
					 else NONE)
			              handle Subscript => #lastCtxt foldState
			              
		     in
		       (Text s, setLastCtxt(foldState, lastCtxt))
		     end
		else let
		       val ctxt = #ctxt treeState
		       val lastCtxt = #lastCtxt foldState
		       val (s',lastCtxt)
			 = List.fold
			   (explode s,
			    ([],lastCtxt),
			    fn (c,(s',lastCtxt))
			     => if Char.isSpace c
				  then case lastCtxt
					 of NONE => (c::s',SOME ctxt)
					  | SOME ctxt'
					  => if Proprec.equals
					        (Proprec.space ctxt, 
						 Proprec.space ctxt')
					       then (s',lastCtxt)
					       else (c::s',SOME ctxt)
				  else (c::s',NONE))
		       val s = implode (List.rev s')
		     in
		       (Text s, setLastCtxt(foldState, lastCtxt))
		     end
	   | (Tagged (t, d), treeState, foldState) 
	   => let
		val treeState = setCtxt(treeState, Proprec.update(#ctxt treeState, t))
		val (d, foldState) = walk(d, treeState, foldState)
	      in 
		(Tagged (t, d), foldState)
	      end

      val (d,_) = walk (d, {ctxt = Proprec.root}, {lastCtxt = NONE})
    in
      d
    end

end