type int = Int.t

functor Synthesize (S : SYNTHESIZE_STRUCTS) : SYNTHESIZE =
struct

open S

structure Color = Document.Color

datatype z = datatype Color.t

type ctxt = Meaning.Proprec.t

type stk = (Document.tag option * Document.item list) list

val initStk : stk = [(NONE,[])]

fun pushChar (stk:stk,c) : stk =
    case stk of 
	(tag,[])::stk => (tag,[Document.Text (String.fromChar c)])::stk
      | (tag,(Document.Text t)::is)::stk =>
	(tag,(Document.Text (t^(String.fromChar c)))::is)::stk
      | (tag,(is as (Document.Tagged (_,_))::_))::stk =>
	(tag,(Document.Text (String.fromChar c))::is)::stk
      | [] => Error.bug "bad stack"
	      
fun pushTag (stk:stk,t) : stk = (SOME t,[])::stk

fun popStk (stk:stk) : stk =
    case stk of
	(SOME t,is)::(t',is')::stk' =>
	(t',(Document.Tagged (t,Document.Doc (List.rev is)))::is')::
	stk'
      | _ => Error.bug "bad stack"

fun popStkN (stk:stk, n:int) =
    if n=0 then stk else popStkN(popStk stk,n-1)

fun unwind stk =
    case stk of
	[(NONE,is)] => Document.Doc (List.rev is)
      | _::_ => unwind (popStk stk)
      | _ => Error.bug "bad stack"

fun synthesize
	(newtags : Document.tag list * ctxt * ctxt -> int * Document.tag list)
	(m : Meaning.t)
	: Document.t =
    let fun loop (i, ctxt, tags, stk) =
	    if i<(Array.length m) then
		let val d = Meaning.toExternal (Array.sub (m,i))
		    val p = #p d
		    val same_ctxt =
			ctxt = p orelse
			(Char.isSpace (#c d) andalso
			 Meaning.Proprec.space ctxt = p)
		in  if same_ctxt then
			loop (i+1, ctxt, tags, pushChar (stk, #c d))
		    else
			let val (remtags,addtags) = newtags (tags,ctxt,p)
			    val stk = popStkN (stk,remtags)
			    val tags' = List.nthTail (tags,remtags)
			    val tags = addtags @ tags'
			    fun aux (t,s) = pushTag(s,t)
			    val stk = List.foldr (addtags,stk,aux)
			in  loop (i+1, p, tags, pushChar (stk, #c d))
			end
		end
	    else
		unwind stk
    in  loop (0, Meaning.Proprec.root, [], initStk)
    end

fun newtags1 (oldtags, ctxt : ctxt, p : ctxt) : int * Document.tag list =
    let val tags = []
	val tags = if #b p then Document.B :: tags else tags
	val tags = if #em p then Document.EM :: tags else tags
	val tags = if #i p then Document.I :: tags else tags
	val tags = if #s p then Document.S :: tags else tags
	val tags = if #tt p then Document.TT :: tags else tags
	fun aux (u,tags) = if u=0 then tags else aux (u-1,Document.U::tags)
	val tags = aux (#u p,tags)
	val tags =
	    case #size p of NONE => tags | SOME s => (Document.Size s)::tags
	val tags =
	    case #color p of Unknown => tags | c => (Document.Color c)::tags
    in  (List.length oldtags, tags)
    end
		 
val synthesize1 =
    synthesize newtags1

end
(*
fun newtags2 (oldtags, ctxt, p) =
    let 
    in
    end

val synthesize2 =
    synthesize newtags2
*)
