structure Main =
struct

type int = Int.t

fun usage s =
   Process.usage {usage = "[-proxy host[:port]] [-mode webequiv] timeLimit",
		  msg = s}

fun parseProxy s =
   case String.tokens (s, fn c => c = #":") of
      [host] => SOME {host = host, port = Port.http}
    | [host, port] =>
	 (case Int.fromString port of
	     NONE => usage (concat ["invalid port: ", port])
	   | SOME p => SOME {host = host, port = p})
    | _ => usage (concat ["invalid -proxy arg: ", s])

val proxy: {host: string, port: int} option ref = ref NONE

fun webequiv args =
   case args of
      [f1, f2] =>
         let
	    val _ = Uri.escapeQuery := false
	    val _ =
	       case !proxy of
		  NONE => 
		     if Net.canPing ("10.1.1.57") then
			proxy := parseProxy ("10.1.1.57:8001")
		     else ()
		| _ => ()
	    val uri =
	       (case Uri.fromString "http://cristal.inria.fr/bin/validate-file"
		   of
		   SOME uri => uri
		 | NONE => Error.bug "bad URL")
            val _ =
	       Net.repeat :=
	       SOME {limit = Time.seconds (IntInf.fromInt 1), tries = 5}
            val ans = In.input (Http.fetch
				{uri = uri,
				 head = false,
				 headers = [],
				 post = SOME (Http.Post.Multipart
					      [("nameone", f1),
					       ("nametwo",f2)]),
				 proxy = !proxy})
         in
            if String.isSubstring {string = ans,
                                   substring = "Error"} orelse
               String.isSubstring {string = ans,
                                   substring = "error"}
               then
                  (Out.output (Out.error, ans)
                   ; raise Fail "webequiv error")
            else if String.isSubstring {string = ans,
                                        substring = "are not equivalent"}
               then
                  (Out.output (Out.error, "documents are not equivalent\n")
                   ; raise Fail "webequiv failed")
            else
                ()
         end
    | _ => usage "webequiv expects two args"

val mode = ref webequiv

fun main args =
   let
      val rest =
	 Popt.parse
	 {switches = args,
	  opts = [("mode",
		   Popt.SpaceString
		   (fn s =>
		    mode := (case s of
                                "webequiv" => webequiv
			      | _ => usage "strange mode"))),
		  ("proxy",
		   Popt.SpaceString (fn p => proxy := parseProxy(p))),
		  Popt.trace]}
    in
      case rest 
	of Result.No s => usage (concat ["invalid switch: ", s])
	 | Result.Yes args => !mode args
   end

fun exportNJ (file: File.t): unit =
   SMLofNJ.exportFn
   (file,
    fn (_, args) => ((main args; OS.Process.success)
		     handle e =>
			let open Layout
			in output (seq [str "main: ", Exn.layout e],
				   Out.error)
			   ; Out.newline Out.error
			   ; OS.Process.failure
			end))

val main = Process.makeMain main

end
